# -*- coding: utf-8 -*-

"""
ACES Configuration Module
This module provides the configuration settings for the ACES project.
"""

from pathlib import Path
import os
import ast
from dotenv import load_dotenv


class Config:
    """
    ACES Configuration Class

    This class contains the configuration settings for the ACES project. These are generated from the .env file.

    Attributes:
        BASEDIR (str): The base directory for data I/O information.
        DATADIR (str): The directory for data collection experiments. DATADIR = BASEDIR / DATADIR
        MODEL_NAME (str): The name of the model.
        MODEL_CHECKPOINT_NAME (str): The name for model checkpoints.
        MODEL_DIR_NAME (str): The name of the model directory. MODEL_DIR = OUTPUT_DIR / MODEL_DIR_NAME
        AUTO_MODEL_DIR_NAME (bool): Flag to use automatic model directory naming.
        # True generates as trial_MODELTYPE + datetime.now() + _v + version
        # False uses the MODEL_DIR_NAME
        FEATURES (str): The list of features used in the model.
        USE_ELEVATION (bool): Flag to use elevation data.
        USE_S1 (bool): Flag to use Sentinel-1 data.
        DERIVE_FEATURES (bool): Flag to derive features.
        ADDED_FEATURES (list): A list of additional features used in the model (used when DERIVE_FEATURES is true).
        LABELS (list): A list of labels used in the model.
        SCALE (int): The scale of the data.
        # Note: The seeding componenet needs fixing. It is not working as expected.
        USE_SEED (bool): Flag to use a seed for reproducbility.
        SEED (int): The seed used for reproducibility.
        PATCH_SHAPE (tuple): The shape of training patches.
        # buffer for prediction purpose
        # Half this will extend on the sides of each patch.
        # if zero; does not do buffer
        # else specify the size as tuple (e.g. 128, 128)
        KERNEL_BUFFER (tuple): The buffer for prediction purpose.
        TRAIN_SIZE (int): The size of the training dataset.
        TEST_SIZE (int): The size of the testing dataset.
        VAL_SIZE (int): The size of the validation dataset.
        BATCH_SIZE (int): The batch size for model training.
        EPOCHS (int): The number of epochs for model training.
        RAMPUP_EPOCHS (int): The number of ramp-up epochs.
        SUSTAIN_EPOCHS (int): The number of sustain epochs.
        USE_ADJUSTED_LR (bool): Flag to use adjusted learning rate.
        MAX_LR (float): The maximum learning rate.
        MID_LR (float): The intermediate learning rate.
        MIN_LR (float): The minimum learning rate.
        DROPOUT_RATE (float): The dropout rate for the model.
        CALLBACK_PARAMETER (str): The parameter used for callbacks.
        # patience for EARLY_STOPPING = int(0.3 * EPOCHS)
        # monitors CALLBACK_PARAMETER
        EARLY_STOPPING (bool): Flag to use early stopping.
        MODEL_TYPE (str): The type of model (current choices: cnn, dnn, unet).
        TRANSFORM_DATA (bool): Flag to transform data (rotation, flip) DNN
        does not do augmentation.
        ACTIVATION_FN (str): The activation function for the model.
        OPTIMIZER (str): The optimizer used for model training.
        # You can use any loss function available [here](https://www.tensorflow.org/api_docs/python/tf/keras/losses):
        # other available are: custom_focal_tversky_loss
        LOSS (str): The loss function used for model training.,
        OUT_CLASS_NUM (int): The number of output classes.
        # This parameter is used in script [5. prediction_unet.py](https://github.com/SERVIR/servir-aces/blob/main/workflow/v2/5.prediction_unet.py)
        # and [host_vertex_ai.py](https://github.com/SERVIR/servir-aces/blob/main/workflow/v2/host_vertex_ai.py) files
        USE_BEST_MODEL_FOR_INFERENCE (bool): Flag to use the best model for inference.
        USE_SERVICE_ACCOUNT (bool): Flag to use a service account for Earth Engine.
        # This is used when USE_SERVICE_ACCOUNT is True
        # This is used to
        EE_SERVICE_CREDENTIALS (str): The path to the Earth Engine service account credentials.
        # Used in the script [5.prediction_unet.py](https://github.com/SERVIR/servir-aces/blob/main/workflow/v2/5.prediction_unet.py) and
        # [5.prediction_dnn](https://github.com/SERVIR/servir-aces/blob/main/workflow/v2/5.prediction_dnn.py)
        EE_OUPUT_ASSET (str): The Earth Engine output (prediction) asset path.
        OUTPUT_NAME (str): The name of the output prediction for GEE asset, locally (in TF Format) and gcs output (in TFRecord format).
        GCS_BUCKET (str): The Google Cloud Storage bucket name.
        # Exported to this dir from [4.export_image_for_prediction.py](https://github.com/SERVIR/servir-aces/blob/main/workflow/v2/4.export_image_for_prediction.py)
        GCS_IMAGE_DIR (str): The Google Cloud Storage image directory to store the image for prediction.
        # used as file_name_prefix = {GCS_IMAGE_DIR}/{GCS_IMAGE_PREFIX} for exporting image
        GCS_IMAGE_PREFIX (str): The Google Cloud Storage image prefix.
        # Used in the script [host_vertex_ai.py](https://github.com/SERVIR/servir-aces/blob/main/workflow/v2/host_vertex_ai.py)
        GCS_PROJECT (str): The Google Cloud Storage project name.
        GCS_VERTEX_MODEL_SAVE_DIR (str): The Google Cloud Storage Vertex AI model save directory.
        GCS_REGION (str): The Google Cloud Storage region.
        GCS_VERTEX_CONTAINER_IMAGE (str): The Google Cloud Storage Vertex AI container image.
        # AI platform expects a need a serialized model, this parameter does this.
        # See its uses in `data_processor.py` and `model_training.py`
        USE_AI_PLATFORM (bool): Flag to use Google Cloud AI Platform.
        # Get machine type here: https://cloud.google.com/vertex-ai/docs/predictions/configure-compute
        GCP_MACHINE_TYPE (str): The Google Cloud Platform machine type.
    """

    def __init__(self, config_file) -> None:

        load_dotenv(config_file)

        self.BASEDIR = Path(os.getenv("BASEDIR"))
        _DATADIR = os.getenv("DATADIR")
        if _DATADIR.startswith("gs://"):
            self.DATADIR = _DATADIR
            self.TRAINING_DIR = f"{self.DATADIR}/training"
            self.TESTING_DIR = f"{self.DATADIR}/testing"
            self.VALIDATION_DIR = f"{self.DATADIR}/validation"
        else:
            self.DATADIR = self.BASEDIR / os.getenv("DATADIR")
            self.TRAINING_DIR = self.DATADIR / "training"
            self.TESTING_DIR = self.DATADIR / "testing"
            self.VALIDATION_DIR= self.DATADIR / "validation"

        print(f"BASEDIR: {self.BASEDIR}")
        print(f"DATADIR: {self.DATADIR}")

        self.OUTPUT_DIR = self.BASEDIR / os.getenv("OUTPUT_DIR")

        self.MODEL_NAME = os.getenv("MODEL_NAME")
        self.MODEL_CHECKPOINT_NAME = os.getenv("MODEL_CHECKPOINT_NAME")

        self.MODEL_DIR_NAME = os.getenv("MODEL_DIR_NAME")

        self.MODEL_DIR = self.OUTPUT_DIR / self.MODEL_DIR_NAME

        self.AUTO_MODEL_DIR_NAME = os.getenv("AUTO_MODEL_DIR_NAME") == "True"
        self.SCALE = int(os.getenv("SCALE"))

        self.FEATURES = os.getenv("FEATURES").split("\n")
        self.ADDED_FEATURES = os.getenv("ADDED_FEATURES").split("\n")
        self.USE_ELEVATION = os.getenv("USE_ELEVATION") == "True"
        self.USE_S1 = os.getenv("USE_S1") == "True"
        self.LABELS = ast.literal_eval(os.getenv("LABELS"))

        if self.USE_ELEVATION:
                self.FEATURES.extend(["elevation", "slope"])

        if self.USE_S1:
            self.FEATURES.extend(["vv_asc_before", "vh_asc_before", "vv_asc_during", "vh_asc_during",
                                  "vv_desc_before", "vh_desc_before", "vv_desc_during", "vh_desc_during"])

        print(f"using features: {self.FEATURES}")
        print(f"using labels: {self.LABELS}")

        self.USE_SEED = os.getenv("USE_SEED") == "True"
        self.SEED = int(os.getenv("SEED"))

        self.PRINT_INFO = os.getenv("USE_SEED") == "True"

        # patch size for training
        self.PATCH_SHAPE = ast.literal_eval(os.getenv("PATCH_SHAPE"))
        self.PATCH_SHAPE_SINGLE = self.PATCH_SHAPE[0]
        KERNEL_BUFFER = os.getenv("KERNEL_BUFFER")
        if KERNEL_BUFFER == "0":
            self.KERNEL_BUFFER = None
        else:
            self.KERNEL_BUFFER = ast.literal_eval(KERNEL_BUFFER)

        # Sizes of the testing, and evaluation datasets
        self.TRAIN_SIZE = int(os.getenv("TRAIN_SIZE"))
        self.TEST_SIZE = int(os.getenv("TEST_SIZE"))
        self.VAL_SIZE = int(os.getenv("VAL_SIZE"))

        self.MODEL_TYPE = os.getenv("MODEL_TYPE")
        self.IS_DNN = True if self.MODEL_TYPE == "dnn" else False

        self.BATCH_SIZE = int(os.getenv("BATCH_SIZE"))
        self.EPOCHS = int(os.getenv("EPOCHS"))
        self.RAMPUP_EPOCHS = int(os.getenv("RAMPUP_EPOCHS"))
        self.SUSTAIN_EPOCHS = int(os.getenv("SUSTAIN_EPOCHS"))

        self.USE_ADJUSTED_LR = os.getenv("USE_ADJUSTED_LR") == "True"
        self.MAX_LR = float(os.getenv("MAX_LR"))
        self.MID_LR = float(os.getenv("MID_LR"))
        self.MIN_LR = float(os.getenv("MIN_LR"))
        self.DROPOUT_RATE = float(os.getenv("DROPOUT_RATE"))

        self.LOSS = os.getenv("LOSS")

        self.OPTIMIZER = os.getenv("OPTIMIZER")

        self.OUT_CLASS_NUM = int(os.getenv("OUT_CLASS_NUM"))

        self.USE_BEST_MODEL_FOR_INFERENCE = os.getenv("USE_BEST_MODEL_FOR_INFERENCE") == "True"

        self.ACTIVATION_FN = "sigmoid" if self.OUT_CLASS_NUM == 1 else "softmax"
        self.CALLBACK_PARAMETER = os.getenv("CALLBACK_PARAMETER")

        self.EARLY_STOPPING = os.getenv("EARLY_STOPPING") == "True"
        self.TRANSFORM_DATA = os.getenv("TRANSFORM_DATA") == "True"
        self.DERIVE_FEATURES = os.getenv("DERIVE_FEATURES") == "True"

        # EE settings
        self.USE_SERVICE_ACCOUNT = os.getenv("USE_SERVICE_ACCOUNT") == "True"
        self.EE_SERVICE_CREDENTIALS = os.getenv("EE_SERVICE_CREDENTIALS")
        self.EE_USER = os.getenv("EE_USER")
        self.EE_OUTPUT_ASSET = os.getenv("EE_OUTPUT_ASSET")
        self.OUTPUT_NAME = os.getenv("OUTPUT_NAME")

        # cloud stuff
        self.GCS_PROJECT = os.getenv("GCS_PROJECT")
        self.GCS_BUCKET = os.getenv("GCS_BUCKET")
        self.GCS_IMAGE_DIR = os.getenv("GCS_IMAGE_DIR")
        self.GCS_IMAGE_PREFIX = os.getenv("GCS_IMAGE_PREFIX")
        self.GCS_VERTEX_MODEL_SAVE_DIR = os.getenv("GCS_VERTEX_MODEL_SAVE_DIR")
        self.GCS_REGION = os.getenv("GCS_REGION")
        self.GCS_VERTEX_CONTAINER_IMAGE = os.getenv("GCS_VERTEX_CONTAINER_IMAGE")

        self.USE_AI_PLATFORM = os.getenv("USE_AI_PLATFORM") == "True"

        self.GCP_MACHINE_TYPE = os.getenv("GCP_MACHINE_TYPE")



# -*- coding: utf-8 -*-

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

try:
    from aces.config import Config
    from aces.model_trainer import ModelTrainer
except ModuleNotFoundError:
    print("ModuleNotFoundError: Attempting to import from parent directory.")
    import os, sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

    from aces.config import Config
    from aces.model_trainer import ModelTrainer



if __name__ == "__main__":
    config_file = "config.env"
    config = Config(config_file)
    trainer = ModelTrainer(config)
    trainer.train_model()



# -*- coding: utf-8 -*-

import logging
logging.basicConfig(level=logging.INFO)

import re
import numpy as np
import matplotlib.pyplot as plt
from typing import List
import tensorflow as tf

__all__ = ["TFUtils", "Utils"]

class TFUtils:
    @staticmethod
    def beam_serialize(patch: np.ndarray) -> bytes:
        features = {
            name: tf.train.Feature(
                float_list=tf.train.FloatList(value=patch[name].flatten())
            )
            for name in patch.dtype.names
        }
        example = tf.train.Example(features=tf.train.Features(feature=features))
        return example.SerializeToString()

    @staticmethod
    def configure_memory_growth() -> List:
        """
        Configure TensorFlow to allocate GPU memory dynamically.

        If GPUs are found, this method enables memory growth for each GPU.
        """
        physical_devices = tf.config.list_physical_devices("GPU")
        # self.config.physical_devices = physical_devices
        if len(physical_devices):
            logging.info(f" > Found {len(physical_devices)} GPUs")
            try:
                for device in physical_devices:
                    tf.config.experimental.set_memory_growth(device, True)
                return physical_devices
            except Exception as err:
                logging.error(err)
        else:
            logging.info(" > No GPUs found")
            return []


class Utils:
    """
    Utils: Utility Functions for ACES

    This class provides utility functions for plotting, splitting data.
    """
    @staticmethod
    def split_dataset(element, num_partitions: int, validation_ratio: float = 0.2, test_ratio: float = 0.2) -> int:
        import random
        weights = [1 - validation_ratio - test_ratio, validation_ratio, test_ratio]
        return random.choices([0, 1, 2], weights)[0]

    @staticmethod
    def plot_metrics(metrics, history, epoch, model_save_dir):
        """
        Plot the training and validation metrics over epochs.

        Args:
            metrics: List of metrics to plot.
            history: Training history containing metric values.
            epoch: Number of epochs.
            model_save_dir: Directory to save the plot.

        Returns:
            None.
        """
        fig, ax = plt.subplots(nrows=len(metrics), sharex=True, figsize=(15, len(metrics) * 6))
        colors = ["#1f77b4", "#ff7f0e", "red", "green", "purple", "orange", "brown", "pink", "gray", "olive", "cyan"]
        for i, metric in enumerate(metrics):
            try:
                ax[i].plot(history[metric], color=colors[i], label=f"Training {metric.upper()}")
                ax[i].plot(history[f"val_{metric}"], linestyle=":", marker="o", markersize=3, color=colors[i], label=f"Validation {metric.upper()}")
                ax[i].set_ylabel(metric.upper())
                ax[i].legend()
            except Exception as e:
                logging.info(f"Exception: {e}")
                logging.info(f"Skipping {metric}.")
                continue

        ax[i].set_xticks(range(1, epoch + 1, 4))
        ax[i].set_xticklabels(range(1, epoch + 1, 4))
        ax[i].set_xlabel("Epoch")
        fig.savefig(f"{model_save_dir}/training.png", dpi=1000)

    @staticmethod
    def filter_good_patches(patch):
        """
        Filter patches to remove those with NaN or infinite values.

        Parameters:
        patch (np.ndarray): The patch to filter.

        Returns:
        bool: True if the patch has no NaN or infinite values, False otherwise.
        """
        # the getdownload url has field names so we"re using view here
        has_nan = np.isnan(np.sum(patch.view(np.float32)))
        has_inf = np.isinf(np.sum(patch.view(np.float32)))
        if has_nan or has_inf:
            return False
        return True

    @staticmethod
    def convert_camel_to_snake(strings):
        converted_strings = []
        for string in strings:
            converted_string = re.sub(r"(?<!^)(?=[A-Z])", "_", string).lower()
            converted_strings.append(converted_string)
        return converted_strings



"""Top-level package for servir-aces."""

from aces.config import *
from aces.data_processor import *
from aces.ee_utils import *
from aces.utils import *
from aces.metrics import *
from aces.model_trainer import *
from aces.model_builder import *
from aces.remote_sensing import *

__author__ = """Biplov Bhandari"""
__email__ = "bionicbiplov45@gmail.com"
__version__ = "0.0.5"



# -*- coding: utf-8 -*-

import logging
logging.basicConfig(level=logging.INFO)

import re, warnings
import ee
import numpy as np
import matplotlib.pyplot as plt
from typing import Union, List
from aces import Config
from aces.utils import Utils
import tensorflow as tf


__all__ = ["EEUtils"]


class EEUtils:
    """
    EEUtils: Earth Engine Utility Class

    This class provides utility functions to handle Earth Engine API information and make authenticated requests.
    """
    @staticmethod
    def get_credentials_by_service_account_key(key):
        """
        Helper function to retrieve credentials using a service account key.

        Parameters:
        key (str): The path to the service account key JSON file.

        Returns:
        ee.ServiceAccountCredentials: The authenticated credentials.
        """
        import json
        service_account = json.load(open(key))
        credentials = ee.ServiceAccountCredentials(service_account["client_email"], key)
        return credentials

    @staticmethod
    def initialize_session(use_highvolume : bool = False, key : Union[str, None] = None):
        """
        Initialize the Earth Engine session.

        Parameters:
        use_highvolume (bool): Whether to use the high-volume Earth Engine API.
        key (str or None): The path to the service account key JSON file. If None, the default credentials will be used.
        """
        if key is None:
            if use_highvolume:
                ee.Initialize(opt_url="https://earthengine-highvolume.googleapis.com")
            else:
                ee.Initialize()
        else:
            credentials = EEUtils.get_credentials_by_service_account_key(key)
            if use_highvolume:
                ee.Initialize(credentials, opt_url="https://earthengine-highvolume.googleapis.com")
            else:
                ee.Initialize(credentials)

    @staticmethod
    def calculate_avg_min_max_statistics(image: ee.Image, geometry: ee.FeatureCollection, scale: int = 30) -> ee.Dictionary:
        """
        Calculate min and max of an image over a specific region.

        Parameters:
        image (ee.Image): The image to calculate statistics on.
        geometry (ee.FeatureCollection): The region to calculate statistics over.
        scale (int, optional): The scale, in meters, of the projection to compute statistics in. Default is 30.

        Returns:
        ee.Dictionary: A dictionary containing the min and max of the image.
        """
        reducers = ee.Reducer.mean() \
            .combine(reducer2=ee.Reducer.min(), sharedInputs=True) \
            .combine(reducer2=ee.Reducer.max(), sharedInputs=True)

        stats = image.reduceRegion(
            reducer=reducers,
            geometry=geometry,
            scale=scale,
            maxPixels=1E13
        )

        return stats

    @staticmethod
    def export_collection_data(collection: ee.FeatureCollection, export_type: Union[list, str]="cloud", start_training=True, **params) -> None:
        if isinstance(export_type, str):
            export_type = [export_type]

        for _type in export_type:
            if _type == "cloud":
                EEUtils._export_collection_to_cloud_storage(collection, start_training, **params)

            if _type == "asset":
                EEUtils._export_collection_to_asset(collection, start_training, **params)

            if _type == "drive":
                EEUtils._export_collection_to_drive(collection, start_training, **params)

            if _type not in ["cloud", "asset", "drive"]:
                raise NotImplementedError(f"Currently supported export types are: {', '.join(_type)}")

    @staticmethod
    def _export_collection_to_drive(collection, start_training, **kwargs) -> None:
        logging.info("Exporting training data to Google Drive..")
        training_task = ee.batch.Export.table.toDrive(
            collection=collection,
            description=kwargs.get("description", "myExportTableTask"),
            folder=kwargs.get("folder", "myFolder"),
            fileNamePrefix=kwargs.get("file_prefix", "myExportTableTask"),
            fileFormat=kwargs.get("file_format", "CSV"),
            selectors=kwargs.get("selectors", collection.first().propertyNames().getInfo()),
        )
        if start_training: training_task.start()

    @staticmethod
    def _export_collection_to_asset(collection, start_training, **kwargs) -> None:
        asset_id = kwargs.get("asset_id", "myAssetId")
        logging.info(f"Exporting training data to {asset_id}..")
        training_task = ee.batch.Export.table.toAsset(
            collection=collection,
            description=kwargs.get("description", "myExportTableTask"),
            assetId=asset_id,
            selectors=kwargs.get("selectors", collection.first().propertyNames().getInfo()),
        )
        if start_training: training_task.start()

    @staticmethod
    def _export_collection_to_cloud_storage(collection, start_training, **kwargs) -> None:
        description = kwargs.get("description", "myExportTableTask")
        bucket = kwargs.get("bucket", "myBucket")
        file_prefix = kwargs.get("file_prefix") if kwargs.get("file_prefix") is not None else description
        logging.info(f"Exporting training data to gs://{bucket}/{file_prefix}..")
        training_task = ee.batch.Export.table.toCloudStorage(
            collection=collection,
            description=description,
            fileNamePrefix=file_prefix,
            bucket=bucket,
            fileFormat=kwargs.get("file_format", "TFRecord"),
            selectors=kwargs.get("selectors", collection.first().propertyNames().getInfo()),
        )
        if start_training: training_task.start()

    @staticmethod
    def beam_export_collection_to_cloud_storage(collection_index, start_training, **kwargs) -> None:
        from aces.ee_utils import EEUtils
        import ee
        EEUtils.initialize_session(use_highvolume=True)

        collection = ee.FeatureCollection(collection_index[0])
        index = collection_index[1]

        description = kwargs.get("description", "myExportTableTask")
        bucket = kwargs.get("bucket", "myBucket")
        file_prefix = kwargs.get("file_prefix") if kwargs.get("file_prefix") is not None else description
        file_prefix = f"{file_prefix}_{index}"
        logging.info(f"Exporting training data to gs://{bucket}/{file_prefix}..")
        training_task = ee.batch.Export.table.toCloudStorage(
            collection=collection,
            description=description,
            fileNamePrefix=file_prefix,
            bucket=bucket,
            fileFormat=kwargs.get("file_format", "TFRecord"),
            selectors=kwargs.get("selectors", collection.first().propertyNames().getInfo()),
        )
        if start_training: training_task.start()

    @staticmethod
    def export_image(image: ee.Image, export_type: str="asset", start_training=True, **params) -> None:
        if isinstance(export_type, str):
            export_type = [export_type]

        for _type in export_type:
            if _type == "cloud":
                EEUtils._export_image_to_cloud_storage(image, start_training, **params)

            if _type == "asset":
                EEUtils._export_image_to_asset(image, start_training, **params)

            if _type not in ["cloud", "asset"]:
                raise NotImplementedError(f"Currently supported export types are: {', '.join(_type)}")

    @staticmethod
    def _export_image_to_cloud_storage(image, start_training, **kwargs) -> None:
        description = kwargs.get("description", "myExportImageTask")
        bucket = kwargs.get("bucket", "myBucket")
        file_name_prefix = kwargs.get("file_name_prefix") if kwargs.get("file_name_prefix") is not None else description
        region = kwargs.get("region", None)
        if region is not None:
            if isinstance(region, ee.FeatureCollection):
                region = region.geometry()
            elif isinstance(region, ee.Geometry):
                region = region
            else:
                raise ValueError(f"region must be an ee.FeatureCollection or ee.Geometry object. Found {type(region)}")

        logging.info(f"Exporting training data to gs://{bucket}/{file_name_prefix}..")

        params = {
            "image": image,
            "description": description,
            "fileNamePrefix": file_name_prefix,
            "bucket": kwargs.get("bucket", "myBucket"),
            "fileFormat": kwargs.get("file_format", "GeoTIFF"),
            "formatOptions": kwargs.get("format_options", None),
            "region": region,
            "scale": kwargs.get("scale", 1000),
            "maxPixels": kwargs.get("max_pixels", 1e10),
        }
        keys = Utils.convert_camel_to_snake(list(params.keys()))
        keys.remove("image")

        for key in list(kwargs.keys()):
            if key not in keys:
                warnings.warn(f"Parameter {key} not found in kwargs. Double check your parameter name (camelCase vs snake_case)")

        not_none_params = {k:v for k, v in params.items() if v is not None}
        training_task = ee.batch.Export.image.toCloudStorage(**not_none_params)
        if start_training: training_task.start()

    @staticmethod
    def _export_image_to_asset(image, start_training, **kwargs) -> None:
        asset_id = kwargs.get("asset_id", "")
        logging.info(f"Exporting image to {asset_id}..")

        training_task = ee.batch.Export.image.toAsset(
            image=image,
            description=kwargs.get("description", "myExportImageTask"),
            assetId=asset_id,
            region=kwargs.get("region", None),
            scale=kwargs.get("scale", 30),
            maxPixels=kwargs.get("max_pixels", 1E13),
        )
        if start_training: training_task.start()

    @staticmethod
    def country_bbox(country_name, max_error=100):
        """Function to get a bounding box geometry of a country

        args:
            country_name (str): US-recognized country name
            max_error (float,optional): The maximum amount of error tolerated when
                performing any necessary reprojection. default = 100

        returns:
            ee.Geometry: geometry of country bounding box
        """

        all_countries = ee.FeatureCollection("USDOS/LSIB_SIMPLE/2017")
        return all_countries.filter(ee.Filter.eq("country_na", country_name))\
                            .geometry(max_error).bounds(max_error)

    @staticmethod
    def get_image_collection_statistics(image_collection: ee.ImageCollection) -> ee.Image:
        reducers = ee.Reducer.mean() \
            .combine(reducer2=ee.Reducer.min(), sharedInputs=True) \
                .combine(reducer2=ee.Reducer.max(), sharedInputs=True) \
                    .combine(reducer2=ee.Reducer.stdDev(), sharedInputs=True) \
                        .combine(reducer2=ee.Reducer.percentile([25, 50, 75], ["Q1", "Q2", "Q3"]), sharedInputs=True)
        reducer = image_collection.reduce(reducer=reducers)
        return reducer.float()

    @staticmethod
    def calculate_planet_indices(image: ee.Image) -> ee.Image:
        ndvi = image.normalizedDifference(["N", "R"]).rename("NDVI")
        evi = image.expression (
            "2.5 * ((NIR - RED) / (NIR + 6 * RED - 7.5 * BLUE + 1))", {
                "NIR": image.select("N"),
                "RED": image.select("R"),
                "BLUE": image.select("B")
            }).rename("EVI")
        ndwi = image.normalizedDifference(["G", "N"]).rename("NDWI")
        savi = image.expression("((NIR - RED) / (NIR + RED + 0.5))*(1.5)", {
            "NIR": image.select("N"),
            "RED": image.select("R")
        }).rename("SAVI")
        msavi2 = image.expression("(( (2*NIR + 1) - sqrt( ((2*NIR + 1) * (2*NIR + 1)) - 8 * (NIR - R) ) )) / 2", {
            "NIR": image.select("N"),
            "R": image.select("R")
        }).rename("MSAVI2")

        mtvi2 = image.expression("( 1.5*(1.2*(NIR - GREEN) - 2.5*(RED - GREEN)) ) / ( sqrt( ((2*NIR + 1) * (2*NIR + 1)) - (6*NIR - 5*sqrt(RED)) - 0.5 ) )", {
            "NIR": image.select("N"),
            "RED": image.select("R"),
            "GREEN": image.select("G"),
        }).rename("MTVI2")

        vari = image.expression("(GREEN - RED) / (GREEN + RED - BLUE)", {
            "GREEN": image.select("G"),
            "RED": image.select("R"),
            "BLUE": image.select("B"),
        }).rename("VARI")

        tgi = image.expression("( (120*(RED - BLUE)) - (190*(RED - GREEN)) ) / 2", {
            "GREEN": image.select("G"),
            "RED": image.select("R"),
            "BLUE": image.select("B"),
        }).rename("TGI")

        return ndvi.addBands([evi, ndwi, savi, msavi2, mtvi2, vari, tgi]).float()

    @staticmethod
    def calculate_evi(image: ee.Image) -> ee.Image:
        evi = image.expression (
            "2.5 * ((NIR - RED) / (NIR + 6 * RED - 7.5 * BLUE + 1))", {
                "NIR": image.select("nir"),
                "RED": image.select("red"),
                "BLUE": image.select("blue")
            }).rename("EVI")
        return evi.float()

    @staticmethod
    def calculate_s1_indices(image: ee.Image) -> ee.Image:
        VV = image.select("VV").rename("vv")
        VH = image.select("VH").rename("vh")
        ratio = VV.divide(VH).rename("s1_ratio")
        ndratio = VV.subtract(VH).divide(VV.add(VH)).rename("s1_ndratio")
        return VV.addBands([VH, ratio, ndratio]).float()

    @staticmethod
    def generate_stratified_samples(image: ee.Image, region: ee.Geometry, numPoints: int = 500, classBand: str = None, scale: int=30, **kwargs) -> ee.FeatureCollection:
        # Add a latitude and longitude band.
        return image.addBands(ee.Image.pixelLonLat()).stratifiedSample(
            numPoints=numPoints,
            classBand=classBand if classBand else "label",
            scale=scale,
            region=region,
            seed=kwargs.get("seed", Config.SEED),
            classValues=kwargs.get("class_values", None),
            classPoints=kwargs.get("class_points", None),# [2000, 600, 600, 600]
        ).map(lambda f: f.setGeometry(ee.Geometry.Point([f.get("longitude"), f.get("latitude")])))

    @staticmethod
    def sample_image_by_collection(image: ee.Image, collection: ee.FeatureCollection, **kwargs: dict) -> ee.FeatureCollection:
        samples = image.sampleRegions(
            collection=collection,
            properties=kwargs.get("properties", collection.first().propertyNames().getInfo()),
            scale=kwargs.get("scale", None),
            geometries=kwargs.get("geometries", False),
            tileScale=kwargs.get("tile_scale", 1),
        )
        return samples

    @staticmethod
    def sample_image(image: ee.Image, region: ee.FeatureCollection, **kwargs: dict) -> ee.FeatureCollection:
        sample = image.sample(region=region,
                              scale=kwargs.get("scale", Config.SCALE),
                              seed=kwargs.get("seed", Config.SEED),
                              geometries=kwargs.get("geometries", False))
        return sample

    @staticmethod
    def beam_yield_sample_points_with_index(index, sample_locations: ee.List, use_service_account: bool = False) -> List:
        from aces.ee_utils import EEUtils
        from aces.config import Config
        import ee
        EEUtils.initialize_session(use_highvolume=True, key=Config.EE_SERVICE_CREDENTIALS if use_service_account else None)
        print(f"Yielding Index: {index} of {sample_locations.size().getInfo() - 1}")
        point = ee.Feature(sample_locations.get(index)).geometry().getInfo()
        return point["coordinates"], index

    @staticmethod
    def beam_yield_sample_points(index, sample_locations: ee.List, use_service_account: bool = False) -> List:
        from aces.ee_utils import EEUtils
        from aces.config import Config
        import ee
        EEUtils.initialize_session(use_highvolume=True, key=Config.EE_SERVICE_CREDENTIALS if use_service_account else None)
        print(f"Yielding Index: {index} of {sample_locations.size().getInfo() - 1}")
        point = ee.Feature(sample_locations.get(index)).geometry().getInfo()
        return point["coordinates"]

    @staticmethod
    def beam_sample_neighbourhood(coords_index, image, use_service_account: bool = False):
        from aces.ee_utils import EEUtils
        from aces.config import Config
        import ee
        EEUtils.initialize_session(use_highvolume=True, key=Config.EE_SERVICE_CREDENTIALS if use_service_account else None)

        coords = coords_index[0]
        index = coords_index[1]

        def get_kernel(kernel_size) -> ee.Kernel:
            eelist = ee.List.repeat(1, kernel_size)
            lists = ee.List.repeat(eelist, kernel_size)
            kernel = ee.Kernel.fixed(kernel_size, kernel_size, lists)
            return kernel

        def create_neighborhood(kernel) -> ee.Image:
            return image.neighborhoodToArray(kernel)

        def sample_data(image, points) -> ee.FeatureCollection:
            return image.sample(
                region=points,
                scale=Config.SCALE,
                tileScale=16,
                geometries=False
            )

        image_kernel = get_kernel(Config.PATCH_SHAPE_SINGLE)
        neighborhood = create_neighborhood(image_kernel)
        training_data = sample_data(neighborhood, ee.Geometry.Point(coords))
        return training_data, index


    @staticmethod
    def beam_get_training_patches(coords: List[float], image: ee.Image, bands: List[str] = [],
                                  scale: int = 5, patch_size: int = 128, use_service_account: bool = False) -> np.ndarray:
        """Get a training patch centered on the coordinates."""
        from aces.ee_utils import EEUtils
        from aces.config import Config
        import ee
        EEUtils.initialize_session(use_highvolume=True, key=Config.EE_SERVICE_CREDENTIALS if use_service_account else None)
        from google.api_core import exceptions, retry
        import requests
        import numpy as np
        from typing import List
        import io

        # @retry.Retry(timeout=10*60) # seconds
        @retry.Retry(deadline=10*60) # seconds
        def get_patch(image: ee.Image, region: ee.Geometry, bands: List[str], patch_size: int) -> np.ndarray:
            """Get the patch of pixels in the geometry as a Numpy array."""
            # Create the URL to download the band values of the patch of pixels.
            url = image.getDownloadURL({
                "region": region,
                "dimensions": [patch_size, patch_size],
                "format": "NPY",
                "bands": bands,
            })
            # Download the pixel data. If we get "429: Too Many Requests" errors,
            # it"s safe to retry the request.
            response = requests.get(url)
            if response.status_code == 429:
                # The retry.Retry library only works with `google.api_core` exceptions.
                raise exceptions.TooManyRequests(response.text)

            if response.status_code == 503:
                raise exceptions.ServiceUnavailable(response.text)

                # Still raise any other exceptions to make sure we got valid data.
            response.raise_for_status()
            # Load the NumPy file data and return it as a NumPy array.
            return np.load(io.BytesIO(response.content), allow_pickle=True)

        # @retry.Retry(timeout=10*60) # seconds
        @retry.Retry(deadline=10*60) # seconds
        def compute_pixel(image: ee.Image, region: ee.Geometry, bands: List[str], patch_size: int, scale_x: float, scale_y: float) -> np.ndarray:
            """Get the patch of pixels in the geometry as a Numpy array."""

            # Make a request object.
            request = {
                "expression": image,
                "fileFormat": "NPY",
                "bandIds": bands,
                "grid": {
                    "dimensions": {
                        "width": patch_size,
                        "height": patch_size
                    },
                    "affineTransform": {
                        "scaleX": scale_x,
                        "shearX": 0,
                        "translateX": coords[0],
                        "shearY": 0,
                        "scaleY": scale_y,
                        "translateY": coords[1]
                    },
                    "crsCode": "EPSG:4326",
                },
            }
            response = ee.data.computePixels(request)
            # Load the NumPy file data and return it as a NumPy array.
            return np.load(io.BytesIO(response.content), allow_pickle=True)

        point = ee.Geometry.Point(coords)
        region = point.buffer(scale * patch_size / 2, 1).bounds(1)
        return get_patch(image, region, bands, patch_size)



# -*- coding: utf-8 -*-

try:
    from aces.ee_utils import EEUtils
except ModuleNotFoundError:
    print("ModuleNotFoundError: Attempting to import from parent directory.")
    import os, sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))

    from aces.ee_utils import EEUtils

try:
    import hydrafloods as hf
except ImportError:
    print("hydrafloods not installed, please install it from pip as `pip install hydrafloods`")

import ee


EEUtils.initialize_session(use_highvolume=True)

country = "Bhutan"
region = ee.FeatureCollection(EEUtils.country_bbox(country))

start = "2021-01-01"
end = "2021-02-01"

# projects/servir-sco-assets/assets/Bhutan/Sentinel1Ascending2021

s1 = hf.Sentinel1(region, start, end)
print(f"number of s1 images: {s1.n_images}")

dems = ee.ImageCollection("COPERNICUS/DEM/GLO30")

print(f"number of dem images: {dems.select('DEM').filterBounds(region).size().getInfo()}")

# Extract out the DEM bands from Copernicus DEM
dem = dems.select("DEM").filterBounds(region).mosaic().unmask(0)

s1_asc = hf.Sentinel1Asc(region, start, end)
# Apply a (psuedo-) terrain flattening algorithm to S1 data
# elevation (ee.Image): Input DEM to calculate slope corrections from
# buffer (int, optional): buffer in meters for layover/shadow mask. If zero then no buffer will be applied. default = 0
s1_asc_flat = s1_asc.apply_func(hf.slope_correction, elevation=dem, buffer=100)

# Apply a speckle filter algorithm to S1 data
s1_asc_filtered = s1_asc_flat.apply_func(hf.refined_lee)
print(f"number of s1_asc_filtered images: {s1_asc_filtered.n_images}")

s1_asc_filtered = s1_asc_filtered.collection.median().clip(region)
s1_asc_filtered = s1_asc_filtered.set("system:time_start", ee.Date(start).millis(), "system:time_end", ee.Date(end).advance(-1, "day").millis())

s1_desc = hf.Sentinel1Desc(region, start, end)
# Apply a (psuedo-) terrain flattening algorithm to S1 data
# elevation (ee.Image): Input DEM to calculate slope corrections from
# buffer (int, optional): buffer in meters for layover/shadow mask. If zero then no buffer will be applied. default = 0
s1_desc_flat = s1_desc.apply_func(hf.slope_correction, elevation=dem, buffer=100)

# Apply a speckle filter algorithm to S1 data
s1_desc_filtered = s1_desc_flat.apply_func(hf.refined_lee)
print(f"number of s1_desc_filtered images: {s1_asc.n_images}")

s1_desc_filtered = s1_desc_filtered.collection.median().clip(region)
s1_desc_filtered = s1_desc_filtered.set("system:time_start", ee.Date(start).millis(), "system:time_end", ee.Date(end).advance(-1, "day").millis());

export_kwargs = {
    "region": region.geometry().bounds().getInfo()["coordinates"],
    "scale": 10,
}

EEUtils.export_image(s1_asc_filtered, export_type="asset", **{**export_kwargs, "asset_id": f"projects/servir-sco-assets/assets/Bhutan/Sentinel1Ascending2021/Ascending_{start}"})
EEUtils.export_image(s1_desc_filtered, export_type="asset", **{**export_kwargs, "asset_id": f"projects/servir-sco-assets/assets/Bhutan/Sentinel1Ascending2021/Descending_{start}"})
